import java.io.*;
import java.net.*;
import java.security.*;
import iaik.security.rsa.*;

				//    Primitive di comunicazione per gli Agenti

				 //  @author	                  Matteo Fasanella

public class commPrimitive {
	
	static int packetLen;
	static byte[] errorB = new byte[1];        // in caso di errore in byte[] torna con byte[0] = -1	
	static byte[] errorT = new byte[1];	//  errore di Time Out
	static int signLen, cryptedPacketLen, simmKeyLen = 0;
	
	
	public commPrimitive () {
		errorB[0] = -1;
		errorT[0] = -77;
	}
	
	public static int send( byte[] packet, InetAddress to,  int port, int localPort, Key key, byte[] sign) {
		
		if (! cryptoPrimitive.verify(Negoziatore.toByteArray(key), sign, key) ) return -1;
		if (! Negoziatore.checkPort(localPort, key) ) return -1;                 // controllo permessi 
		
		DatagramSocket socket = null;
		DatagramPacket dgram;
		try {
			socket = new DatagramSocket(localPort);
			dgram = new DatagramPacket (packet, packet.length, to, port);
			socket.send ( dgram);
			socket.close();
			return 0;
		} catch (Exception e) { return -1;}
	}
	
	public static byte[] recv(int port, int  timeOut, Key key, byte[] sign) {
		byte[] errore = new byte[1];
		errore[0] = -1;
		
		if (! cryptoPrimitive.verify(Negoziatore.toByteArray(key), sign, key) ) return errore;
		if (! Negoziatore.checkPort(port, key) ) return errore;  // controlla se la porta  stata assegnata a lui....
		
		DatagramSocket socket = null;
		DatagramPacket dgram;
		try {
			socket = new DatagramSocket();
			dgram = new DatagramPacket (new byte[packetLen], packetLen);
			
			socket.setSoTimeout(timeOut/1000);
			socket.receive ( dgram);
			
			socket.close();
			
			return dgram.getData();
		} catch (InterruptedIOException ex) { return errorT; 
		} catch (Exception e) { return errorB;}
	}
	
	public static int getPort(Key key, byte[] sign) {
	
		if (! cryptoPrimitive.verify(Negoziatore.toByteArray(key), sign, key) ) return -1;		
		return Negoziatore.getFreePort(key, sign);
	}
	
	
	public static int sendMBox ( byte[] packet, InetAddress to, int alias, int localPort, int TOut, Key key, byte[] sign,  int myAlias)  {

		Key simmKey = cryptoPrimitive.createKey();

		if (! cryptoPrimitive.verify(Negoziatore.toByteArray(key), sign, key) ) return -1;
		if (! Negoziatore.checkPort(localPort, key) ) return -1;                      // controlla i permessi....
		
		byte[] negPacket = makeAliasPacket(key, sign, simmKey, packet.length, alias, to, localPort, myAlias);  // crea il pacchetto di negoziazione
		send (negPacket, to, 55555, localPort, key, sign);
		
		byte[] portrecv = new byte[3];
		portrecv = recv (localPort, TOut, key, sign);
		if ( portrecv[0] == -1) return -22;   
		int remPort = portrecv[0] * 10000  +  portrecv[1]*100 + portrecv[2];   //  per es 55555   5|55|55
		int ret = send (packet, to, remPort, localPort, key, sign);
		if (ret == -1 ) return -22;
		return 0;
	}

	
	public static byte[] makeNegPacket (Key key, byte[] sign, Key simmKey,  int port, InetAddress to) {
		byte[] ritorno = new byte[2048];
		byte[] cryptedRitorno = new byte[cryptedPacketLen];
   		byte[] serializedKey = new byte[789];		
   		byte[] serializedSimmKey = new byte[789];   		
		
	    try {	
   		ByteArrayOutputStream baos = new ByteArrayOutputStream();
		ObjectOutputStream oos = new ObjectOutputStream (baos);
		oos.writeObject (key);
		serializedKey = baos.toByteArray();
		
   		ByteArrayOutputStream baos2 = new ByteArrayOutputStream();
		ObjectOutputStream oos2 = new ObjectOutputStream (baos2);
		oos2.writeObject (key);
		serializedSimmKey = baos2.toByteArray();
	  } catch (Exception e) { System.out.println("Exception: "+ e.toString()); ritorno[0] = -1; return ritorno; }

		//  crea il pacchetto mettendo i vari elementi nel vettore di byte ritorno, nei posti giusti

		for (int i=0; i<=788; i++) ritorno[i] = serializedKey[i];
		for (int i=788; i<=788+signLen; i++) ritorno[i] = sign[i-789];
		for (int i=788+signLen; i<=788+signLen+4; i++) ritorno[i] = 0;		// Flag				
		ritorno[788+signLen+7] = 0;							// port  ( 4 bytes)
		ritorno[788+signLen+8] = (byte) (port %100);	port = port /100;
		ritorno[788+signLen+9] = (byte) (port %100);	port = port /100;	
		ritorno[788+signLen+10] = (byte) port;
		for (int i=2047-simmKeyLen; i<=2047; i++) ritorno[i] = serializedSimmKey[i-(2047-simmKeyLen)];
		
		if (to == null) return ritorno;
		RSAPublicKey serverKey = (RSAPublicKey) getServerKey(to);
		if (serverKey == null) { ritorno[0] =-1;  return ritorno; }
		cryptedRitorno = cryptoPrimitive.cypher( ritorno, serverKey);
		
		return cryptedRitorno;
}


	public static byte[] makeAliasPacket (Key key, byte[] sign, Key simmKey, int packetLen, int alias, InetAddress to, int port, int myAlias) {
		byte[] ritorno = new byte[2048];
		byte[] cryptedRitorno = new byte[cryptedPacketLen];		
		
		ritorno = makeNegPacket (key, sign, simmKey,  port, null);
		
		ritorno[788+signLen] = (byte) (alias % 100);		alias =alias / 100;
		ritorno[788+signLen+1] = (byte) (alias % 100);		alias =alias / 100;
		ritorno[788+signLen+2] = (byte) (alias % 100);		alias =alias / 100;
		ritorno[788+signLen+3] = (byte) (alias);		
		
		ritorno[788+signLen+4] = (byte) ( packetLen/1024 + (1 - 1 * (packetLen%1024))  );
		
		byte[] thisHost = getHostAddress ();
		ritorno[788+signLen+5] = thisHost[1];
		ritorno[788+signLen+6] = thisHost[2];
		ritorno[788+signLen+7] = thisHost[3];
		ritorno[788+signLen+8] = thisHost[4];		
		
		if (myAlias !=0) {
			ritorno[788+signLen+9] = (byte) (myAlias % 100);		myAlias = myAlias / 100;
			ritorno[788+signLen+10] = (byte) (myAlias % 100);		myAlias = myAlias / 100;
			ritorno[788+signLen+11] = (byte) (myAlias % 100);		myAlias = myAlias / 100;
			ritorno[788+signLen+12] = (byte) (myAlias % 100);		
			}

		RSAPublicKey serverKey = (RSAPublicKey) getServerKey(to);
		if (serverKey == null) { ritorno[0] =-1;  return ritorno; }
		cryptedRitorno = cryptoPrimitive.cypher( ritorno, serverKey);
		
		return cryptedRitorno;
		
	}
	

	public static byte[] makeNextAliasPacket (Key key, byte[] sign, Key simmKey, int packetLen, int alias, int nextAlias,InetAddress to, int port, int myAlias) {
		byte[] ritorno = new byte[2048];
		byte[] cryptedRitorno = new byte[cryptedPacketLen];		
		
		ritorno = makeNegPacket (key, sign, simmKey,  port, null);
		
		ritorno[788+signLen] = (byte) (alias % 100);		alias =alias / 100;
		ritorno[788+signLen+1] = (byte) (alias % 100);		alias =alias / 100;
		ritorno[788+signLen+2] = (byte) (alias % 100);		alias =alias / 100;
		ritorno[788+signLen+3] = (byte) (alias);		
		
		ritorno[788+signLen+4] = (byte) ( packetLen/1024 + (1 - 1 * (packetLen%1024))  );
		
		byte[] thisHost = getHostAddress ();
		ritorno[788+signLen+5] = thisHost[1];
		ritorno[788+signLen+6] = thisHost[2];
		ritorno[788+signLen+7] = thisHost[3];
		ritorno[788+signLen+8] = thisHost[4];		
		
		if (nextAlias !=0) {
			ritorno[788+signLen+17] = (byte) (nextAlias % 100);		nextAlias = nextAlias / 100;
			ritorno[788+signLen+18] = (byte) (nextAlias % 100);		nextAlias = nextAlias / 100;
			ritorno[788+signLen+19] = (byte) (nextAlias % 100);		nextAlias = nextAlias / 100;
			ritorno[788+signLen+20] = (byte) (nextAlias % 100);		
			}

		if (myAlias !=0) {
			ritorno[788+signLen+9] = (byte) (myAlias % 100);		myAlias = myAlias / 100;
			ritorno[788+signLen+10] = (byte) (myAlias % 100);		myAlias = myAlias / 100;
			ritorno[788+signLen+11] = (byte) (myAlias % 100);		myAlias = myAlias / 100;
			ritorno[788+signLen+12] = (byte) (myAlias % 100);		
			}


		RSAPublicKey serverKey = (RSAPublicKey) getServerKey(to);
		if (serverKey == null) { ritorno[0] =-1;  return ritorno; }
		cryptedRitorno = cryptoPrimitive.cypher( ritorno, serverKey);
		
		return cryptedRitorno;
		
	}



	public static Key getServerKey( InetAddress server) {
		
		// cerca nel file Server.key la chiave corrispondente all'InetAddress server ( o in una tabella Hash)
		
		// se non la trova la chiede al server stesso
		
		// altrimenti ritorna null
		
		return null;
	}
	
	private static byte[] getHostAddress()  {
		byte[] ritorno = new byte[4];
		ritorno[0] = -1;
		
		try { 	// legge l'indirizzo dell'host corrente
			InetAddress hostAddress = InetAddress.getByName("localhost");
		} catch (Exception e) { System.out.println("Exception: " + e.toString() ); return ritorno;}
		
		// lo converte in byte[4] e lo scrive in ritorno
		
		return ritorno;
	}

}